from fastapi import FastAPI 
from text_preprocessing import text_preprocessing 
from ner import ner_spacy,find_imo,find_swift,ner_dicts
from highlight_pdf import output,base64_to_pdf,pdf_to_base64
app = FastAPI()
# Define the default route 
@app.get("/")
def root():
    return {"message": "Welcome to NER FastAPI"}

@app.get("/entities")
def get_entites(text):
    """
    Text is the text from wich we extract the entities
    """ 
    preprocessed_text = text_preprocessing(text)
    ents = list(ner_spacy(preprocessed_text)[0])
    ents.extend(find_imo(text))
    ents.extend(find_swift(text))
    ents.extend(ner_dicts(text,"postgresql://postgres:achraf@localhost:5432/ner_dicts","ports_banks_ships"))
    return {"text":text,"entities":[(ent.text,ent.label_) for ent in ents]}
@app.get("/highlighted_pdf")
def highlight_pdf(pdf:str):
    pdf =base64_to_pdf(pdf)
    highlighted_pdf,positions = output(pdf)
    highlighted_pdf_base64 = pdf_to_base64(highlighted_pdf)
    # positions format: a dict containing entities as keys, and the values a list of tuples:(list of Rect positions,page number)
    return {"highlighted pdf":highlighted_pdf_base64,"entities":positions}





from text_preprocessing import spacy_preprocessing, detect
import spacy
from spaczz.pipeline import SpaczzRuler
import pandas as pd
from sqlalchemy import create_engine
def ner_spacy(text):
    ner  = spacy.load("en_core_web_sm")
    text = spacy_preprocessing(text,lowercase=True,lemmatize=True)
    labels = ner.get_pipe("ner").labels
    return ner(text).ents,labels




def find_swift(text):
    if detect(text) == "en":
        nlp  = spacy.load("en_core_web_sm",disable=["tagger","parser","ner"])
    else:
        nlp  = spacy.load("fr_core_news_sm",disable=["tagger","parser","ner"])
    if "entity_ruler" not in nlp.pipe_names:
        ruler = nlp.add_pipe("entity_ruler",after="ner")
    else:
        ruler= nlp.get_pipe("entity_ruler")
    text = spacy_preprocessing(text,lowercase=True,lemmatize=True)
    # we use spacy s entity ruler for matching
    patterns =[
        ## uppercases will be lower after preprocessing
         #matches :code(optional) all words wich their lower case form is swift followed  by  space or punct and the code
        {"label":"Code Swift","pattern":[{"LOWER":"code","OP":"?"},{"LOWER":"swift"},{'IS_SPACE':True,'OP':'?'},{"IS_PUNCT":True,"OP":"?"},{"TEXT": {"REGEX": "[A-Za-z0-9]{8,11}"}}]},
        #matches : Swift(e)(with some possible typos) tied to the code
        {"label":"Code Swift","pattern":[{"TEXT": {"REGEX": "(swift)[e]?[A-Za-z0-9]{8,11}"}}]},
        #matches abbreviations:
        {"label":"Code Swift","pattern":[{"TEXT": {"REGEX": "^(sft|st)[e]?[A-Za-z0-9]{8,11}"}}]},
        # matches swift(any special character)code
        {"label":"Code Swift","pattern":[{"TEXT": {"REGEX": "^(swift|sft|st)[e]?[!@#$%^&*()_+\-=\[\]{};':'\\|,.<>\/?][A-Za-z0-9]{8,11}"}}]}
        ]
    ruler.add_patterns(patterns)
    tokens = nlp(text)
    return tokens.ents


def check_imo_funct(imo_code):
    if len(str(imo_code))!=7:
        return False
    else:
        digits = [int(x) for x in str(imo_code)]
        s=0
        for i in range(len(digits)-1):
            s+= digits[i]*(len(digits)-i)
        if s%10==digits[6]:
            return True
        else :
            return False

def find_imo(text):
    if detect(text) == "en":
        nlp = spacy.load("en_core_web_sm",disable=["tagger","parser","ner"])
    else:
        nlp  = spacy.load("fr_core_news_sm",disable=["tagger","parser","ner"])    
    if "entity_ruler" not in nlp.pipe_names:
        ruler = nlp.add_pipe("entity_ruler")
    else:
        ruler= nlp.get_pipe("entity_ruler")
    text = spacy_preprocessing(text)
    # we use spacy s entity ruler for matching
    patterns =[
        #matches : imo 2313... or imo : 2313... ...
        {"label":"Code IMO","pattern":[{"LOWER":"code","OP":"?"},{"LOWER":"imo"},{'IS_SPACE':True,'OP':'?'},{"IS_PUNCT":True,"OP":"?"},{"TEXT": {"REGEX": "[0-9]{7}"}}]},
        #matches imo2343..., also 2313...(alone)
        {"label":"Code IMO","pattern":[{"TEXT": {"REGEX": "^(IMO|imo)?[0-9]{7}"}}]},
         # matches imo(possible typo)(any special character)code
        {"label":"Code IMO","pattern":[{"TEXT": {"REGEX": "^(imo)[a-z]?[!@#$%^&*()_+\-=\[\]{};':'\\|,.<>\/?][0-9]{7}"}}]}
        ]
    ruler.add_patterns(patterns)
    tokens = nlp(text)
    l=[ent[-1] for ent in tokens.ents]
    # we try to extract only the number to check if its really an imo
    possible_imos={}
    for ent,elt in zip(tokens.ents,l) :
        num= str(elt)[-7:]
        possible_imos[ent]=elt
    #check if imo:
    imos = []
    for ent,num in possible_imos.items():
        if check_imo_funct(num):
            imos.append(ent)
    return imos
#ner with dicts
def fuzzy_ner(text,keywords:dict):
    if detect(text) == "en":
        nlp = spacy.load("en_core_web_sm",disable=["tagger","parser","ner"])
    else:
        nlp  = spacy.load("fr_core_news_sm",disable=["tagger","parser","ner"])

    doc  = nlp(text) 
    patterns=[]
    for key,value in keywords.items():
            if type(value)==list:
                for v in value:
                    patterns.append({"label":key,"pattern":v,"type":"fuzzy"})
            else:
                patterns.append({"label":key,"pattern":value,"type":"fuzzy"})

    ruler = SpaczzRuler(nlp)
    ruler.add_patterns(patterns)
    doc = ruler(doc)
    return doc.ents

def ner_dicts(text,db,table):
    """
    db format : 'postgresql://user:password@host:port/db'
    """
    engine = create_engine(db)
    df = pd.read_sql("SELECT * FROM {}".format(table),con=engine)
    df.drop("index",axis=1,inplace=True)
    df = df.reindex(columns=["Type","name"])
    entity_list=dict(df.values)
    return fuzzy_ner(text,entity_list)
    

    


# Import Libraries
import fitz
from ner import ner_spacy , find_imo,find_swift,ner_dicts
import base64
#extract text from pdf
def extract_text(input_file):
    doc = fitz.open(input_file)
    text = []
    for page in doc:
        text.append(page.get_text("text"))
    return text

#find entities
def find_ent(input_file):
    text_ents  =[]
    for txt in  extract_text(input_file):
        page_ents = []
        page_ents.extend(ner_spacy(txt)[0]) #ner_spacy(txt)[0]:entities
        page_ents.extend(find_imo(txt))
        page_ents.extend(find_swift(txt))
        page_ents.extend(ner_dicts(txt,"postgresql://postgres:achraf@localhost:5432/ner_dicts","ports_banks_ships"))
        text_ents.append(page_ents)
    return text_ents

#highlight entities in pages
def highlight_ent(page ,page_no, matching_ents):
    positions_dict = {}
    for ent in matching_ents:
        matching_val_area = page.search_for(ent.text)
        positions_dict[ent] =(matching_val_area,page_no)
        highlight = page.addHighlightAnnot(matching_val_area)
        info = highlight.info
        info["title"] = ent.label_
        info["content"] = ent.label_
        highlight.set_info(info)
        highlight.update()
    return positions_dict
#pdf to base64
def pdf_to_base64(pdf):
    with open(pdf, "rb") as pdf_file:
        encoded_string = base64.b64encode(pdf_file.read())
    return encoded_string
#base64 to pdf
def base64_to_pdf(base64):
    with open('out.pdf', 'wb') as pdf:
        pdf.write(base64.b64decode(base64))
    return ".//out.pdf"
#final function
def output(input_file):
    doc=fitz.open(input_file)
    text_ents = find_ent(input_file)
    i=0
    positions = {}
    for page in doc:
        for ent in text_ents[i]:
            if ent not in positions.keys():
                positions[ent] =[]
        part_pos=highlight_ent(page,i+1,text_ents[i])
        i+=1
        for entity,pos in part_pos.items():
            positions[entity].append(pos)
    doc.save("output.pdf", garbage=4, deflate=True, clean=True)
    return ".//ouput.pdf",positions


# Text Preprocessing with NLTK
#import libraries
import nltk
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import WhitespaceTokenizer , TreebankWordTokenizer,WordPunctTokenizer,word_tokenize
from nltk.corpus import stopwords
import string
import contractions
from spellchecker import SpellChecker
import re
from french_lefff_lemmatizer.french_lefff_lemmatizer import FrenchLefffLemmatizer
import unicodedata
from langdetect import detect
from itertools import chain , combinations
nltk.download('stopwords')
nltk.download('punkt')
nltk.download('words')
nltk.download('wordnet')

def remove_urls(text):
    url_pattern = re.compile(r'https?://\S+|www\.\S+')
    return url_pattern.sub(r'', text)

def remove_accented_chars(text):
    new_text = unicodedata.normalize('NFKD', text).encode('ascii', 'ignore').decode('utf-8', 'ignore')
    return new_text


def text_preprocessing(text,accented=False,stopw=False,punctuation=False,lowercase=False,lemmatize=False,spelling=False,expand_contraction=False,urls=False):
    if detect(text)=='en':
        stopword =stopwords.words('english')
        lemmatizer = WordNetLemmatizer()
        spell = SpellChecker()
    else :#if lang="french"
        stopword = stopwords.words('french')
        lemmatizer = FrenchLefffLemmatizer()
        spell = SpellChecker(language="fr")
    if lowercase:
        #lowercase the text 
        text = text.lower()
    if urls:
        #remove urls
        text=remove_urls(text)
    #tokenize the text 
    tokens =WhitespaceTokenizer().tokenize(text)
    if expand_contraction:
        #expand contractions
        tokens = [contractions.fix(token) for token in tokens]
    if punctuation:
        #remove punctuation
        tokens = [token for token in tokens if token not in string.punctuation]
    if stopw:
        #remove stopwords
        tokens = [token for token in tokens if token not in stopword]
    if accented:
        tokens = [remove_accented_chars(token) for token in tokens]
    if spelling:
        #spell check:
        tokens = [spell.correction(token) for token in tokens]
    if lemmatize:
        #lemmatization : 
        tokens = [lemmatizer.lemmatize(token) for token in tokens]
    return tokens

    #Some tests:
file = 'C:/Users/PC2/Downloads/test-ex.txt'
f=open(file,'r')
data = f.read()
#print(text_preprocessing(data))


# Text Preprocessing with spaCy

#Now we ll do preprocessing using mainly spacy
import spacy
#load only french and english models tokenizers
nlp_en = spacy.load("en_core_web_sm", disable=['parser', 'tagger', 'ner'])
nlp_fr = spacy.load("fr_core_news_sm", disable=['parser', 'tagger', 'ner'])

def spacy_preprocessing(text,lowercase=True,stopw=True,punctuation=True,alphabetic=True,lemmatize=True,):
    if detect(text)=="en":
        nlp = nlp_en
    else :
        nlp = nlp_fr
    if lowercase:
        text = text.lower()
    remove_accented_chars(text)
    #tokenize with spacy's default tokenizer
    tokens = nlp(text)
    if stopw :
        tokens = [token for token in tokens if not token.is_stop]
    if lemmatize :
        tokens = [token.lemma_.strip() for token in tokens]
    if punctuation :
        tokens = [re.sub('<[^>]*>', '', token) for token in tokens]
    if alphabetic:
        tokens = [re.sub('[\W]+','',token) for token in tokens]
    return ' '.join(word for word in tokens)

file = 'C:/Users/PC2/Downloads/test-ex.txt'
f=open(file,'r')
data = f.read()
#print(spacy_preprocessing(data,lowercase=False))








